@@ -19,7 +19,7 @@ class CloudFlare(object):
     class _v4base(object):
         """ Cloudflare v4 API"""
 
-        def __init__(self, email, token, certtoken, base_url, debug, raw):
+        def __init__(self, email, token, certtoken, base_url, debug, raw, use_sessions):
             """ Cloudflare v4 API"""
 
             self.email = email
@@ -27,6 +27,8 @@ class CloudFlare(object):
             self.certtoken = certtoken
             self.base_url = base_url
             self.raw = raw
+            self.use_sessions = use_sessions
+            self.session = None
             self.user_agent = user_agent()
 
             if debug:
@@ -47,6 +49,14 @@ class CloudFlare(object):
                               identifier1, identifier2, identifier3,
                               params, data, files)
 
+        def _add_auth_headers(self, headers):
+            """ Add authentication headers """
+            if self.email:
+                headers['X-Auth-Email'] = self.email
+                headers['X-Auth-Key'] = self.token
+            else:
+                headers['Authorization'] = 'Bearer {}'.format(self.token)
+
         def call_with_auth(self, method, parts,
                            identifier1=None, identifier2=None, identifier3=None,
                            params=None, data=None, files=None):
@@ -56,10 +66,12 @@ class CloudFlare(object):
                 raise CloudFlareAPIError(0, 'no email and/or token defined')
             headers = {
                 'User-Agent': self.user_agent,
-                'X-Auth-Email': self.email,
-                'X-Auth-Key': self.token,
                 'Content-Type': 'application/json'
             }
+            self._add_auth_headers(headers)
+            if type(data) == str:
+                # passing javascript vs JSON
+                headers['Content-Type'] = 'application/javascript'
             if files:
                 # overwrite Content-Type as we are uploading data
                 headers['Content-Type'] = 'multipart/form-data'
@@ -69,6 +81,30 @@ class CloudFlare(object):
                               identifier1, identifier2, identifier3,
                               params, data, files)
 
+        def call_with_auth_unwrapped(self, method, parts,
+                                     identifier1=None, identifier2=None, identifier3=None,
+                                     params=None, data=None, files=None):
+            """ Cloudflare v4 API"""
+
+            if self.email is '' or self.token is '':
+                raise CloudFlareAPIError(0, 'no email and/or token defined')
+            headers = {
+                'User-Agent': self.user_agent,
+                'Content-Type': 'application/json'
+            }
+            self._add_auth_headers(headers)
+            if type(data) == str:
+                # passing javascript vs JSON
+                headers['Content-Type'] = 'application/javascript'
+            if files:
+                # overwrite Content-Type as we are uploading data
+                headers['Content-Type'] = 'multipart/form-data'
+                # however something isn't right and this works ... look at again later!
+                del headers['Content-Type']
+            return self._call_unwrapped(method, headers, parts,
+                                        identifier1, identifier2, identifier3,
+                                        params, data, files)
+
         def call_with_certauth(self, method, parts,
                                identifier1=None, identifier2=None, identifier3=None,
                                params=None, data=None, files=None):
@@ -85,22 +121,24 @@ class CloudFlare(object):
                               identifier1, identifier2, identifier3,
                               params, data, files)
 
-        def _raw(self, method, headers, parts,
-                 identifier1=None, identifier2=None, identifier3=None,
-                 params=None, data=None, files=None):
+        def _network(self, method, headers, parts,
+                     identifier1=None, identifier2=None, identifier3=None,
+                     params=None, data=None, files=None):
             """ Cloudflare v4 API"""
 
             if self.logger:
-                self.logger.debug('Call: %s,%s,%s,%s,%s,%s' % (str(parts[0]),
-                                                               str(identifier1),
-                                                               str(parts[1]),
-                                                               str(identifier2),
-                                                               str(parts[2]),
-                                                               str(identifier3)))
-                self.logger.debug('Call: optional params and data %s %s' % (str(params),
-                                                                            str(data)))
+                self.logger.debug('Call: %s,%s,%s,%s,%s,%s',
+                                  str(parts[0]),
+                                  str(identifier1),
+                                  str(parts[1]),
+                                  str(identifier2),
+                                  str(parts[2]),
+                                  str(identifier3))
+                self.logger.debug('Call: optional params and data %s %s',
+                                  str(params),
+                                  str(data))
                 if files:
-                    self.logger.debug('Call: upload file %r' % (files))
+                    self.logger.debug('Call: upload file %r', files)
 
             if (method is None) or (parts[0] is None):
                 # should never happen
@@ -134,26 +172,72 @@ class CloudFlare(object):
                 url += '/' + identifier3
 
             if self.logger:
-                self.logger.debug('Call: method and url %s %s' % (str(method), str(url)))
-                self.logger.debug('Call: headers %s' % str(sanitize_secrets(headers)))
+                self.logger.debug('Call: method and url %s %s', str(method), str(url))
+                self.logger.debug('Call: headers %s', str(sanitize_secrets(headers)))
 
             method = method.upper()
 
             if self.logger:
                 self.logger.debug('Call: doit!')
 
+            if self.use_sessions:
+                if self.session is None:
+                    self.session = requests.Session()
+            else:
+                self.session = requests
+
             try:
                 if method == 'GET':
-                    response = requests.get(url, headers=headers, params=params, data=data)
+                    response = self.session.get(url,
+                                                headers=headers,
+                                                params=params,
+                                                data=data)
                 elif method == 'POST':
-                    response = requests.post(url, headers=headers, params=params, json=data, files=files)
+                    if type(data) == str:
+                        response = self.session.post(url,
+                                                     headers=headers,
+                                                     params=params,
+                                                     data=data,
+                                                     files=files)
+                    else:
+                        response = self.session.post(url,
+                                                     headers=headers,
+                                                     params=params,
+                                                     json=data,
+                                                     files=files)
                 elif method == 'PUT':
-                    response = requests.put(url, headers=headers, params=params, json=data)
+                    if type(data) == str:
+                        response = self.session.put(url,
+                                                    headers=headers,
+                                                    params=params,
+                                                    data=data)
+                    else:
+                        response = self.session.put(url,
+                                                    headers=headers,
+                                                    params=params,
+                                                    json=data)
                 elif method == 'DELETE':
-                    response = requests.delete(url, headers=headers, json=data)
+                    if type(data) == str:
+                        response = self.session.delete(url,
+                                                       headers=headers,
+                                                       params=params,
+                                                       data=data)
+                    else:
+                        response = self.session.delete(url,
+                                                       headers=headers,
+                                                       params=params,
+                                                       json=data)
                 elif method == 'PATCH':
-                    response = requests.request('PATCH', url,
-                                                headers=headers, params=params, json=data)
+                    if type(data) == str:
+                        response = self.session.request('PATCH', url,
+                                                        headers=headers,
+                                                        params=params,
+                                                        data=data)
+                    else:
+                        response = self.session.request('PATCH', url,
+                                                        headers=headers,
+                                                        params=params,
+                                                        json=data)
                 else:
                     # should never happen
                     raise CloudFlareAPIError(0, 'method not supported')
@@ -178,10 +262,13 @@ class CloudFlare(object):
                 # API should always response; but if it doesn't; here's the default
                 response_type = 'application/octet-stream'
             response_code = response.status_code
-            response_data = response.text
+            response_data = response.content
+            if type(response_data) != str:
+                response_data = response_data.decode("utf-8")
 
             if self.logger:
-                self.logger.debug('Response: %d, %s, %s' % (response_code, response_type, response_data))
+                self.logger.debug('Response: %d, %s, %s',
+                                  response_code, response_type, response_data)
 
             if response_code >= 500 and response_code <= 599:
                 # 500 Internal Server Error
@@ -223,13 +310,52 @@ class CloudFlare(object):
             #
             # should be a 200 response at this point
 
+            return [response_type, response_code, response_data]
+
+        def _raw(self, method, headers, parts,
+                 identifier1=None, identifier2=None, identifier3=None,
+                 params=None, data=None, files=None):
+            """ Cloudflare v4 API"""
+
+            [response_type, response_code, response_data] = self._network(method,
+                                                                          headers, parts,
+                                                                          identifier1,
+                                                                          identifier2,
+                                                                          identifier3,
+                                                                          params, data, files)
+
             if response_type == 'application/json':
                 # API says it's JSON; so it better be parsable as JSON
+                # NDJSON is returned by Enterprise Log Share i.e. /zones/:id/logs/received
+                if hasattr(response_data, 'decode'):
+                    response_data = response_data.decode('utf-8')
                 try:
                     response_data = json.loads(response_data)
                 except ValueError:
-                    # While this should not happen; it's always possible
-                    raise CloudFlareAPIError(0, 'JSON parse failed - report to Cloudflare.')
+                    if response_data == '':
+                        # This should really be 'null' but it isn't. Even then, it's wrong!
+                        if response_code == requests.codes.ok:
+                            # 200 ok
+                            response_data = {'success': True,
+                                             'result': None}
+                        else:
+                            # 3xx & 4xx errors
+                            response_data = {'success': False,
+                                             'code': response_code,
+                                             'result': None}
+                    else:
+                        # Lets see if it's NDJSON data
+                        # NDJSON is a series of JSON elements with newlines between each element
+                        try:
+                            r = []
+                            for l in response_data.splitlines():
+                                r.append(json.loads(l))
+                            response_data = r
+                        except:
+                            # While this should not happen; it's always possible
+                            if self.logger:
+                                self.logger.debug('Response data not JSON: %r', response_data)
+                            raise CloudFlareAPIError(0, 'JSON parse failed - report to Cloudflare.')
 
                 if response_code == requests.codes.ok:
                     # 200 ok - so nothing needs to be done
@@ -240,6 +366,8 @@ class CloudFlare(object):
                     pass
             elif response_type == 'text/plain' or response_type == 'application/octet-stream':
                 # API says it's text; but maybe it's actually JSON? - should be fixed in API
+                if hasattr(response_data, 'decode'):
+                    response_data = response_data.decode('utf-8')
                 try:
                     response_data = json.loads(response_data)
                 except ValueError:
@@ -250,16 +378,44 @@ class CloudFlare(object):
                         response_data = {'success': True, 'result': str(response_data)}
                     else:
                         # 3xx & 4xx errors
-                        response_data = {'success': False, 'code': response_code, 'result': str(response_data)}
+                        response_data = {'success': False,
+                                         'code': response_code,
+                                         'result': str(response_data)}
+            elif response_type == 'text/javascript' or response_type == 'application/javascript':
+                # used by Cloudflare workers
+                if response_code == requests.codes.ok:
+                    # 200 ok
+                    response_data = {'success': True,
+                                     'result': str(response_data)}
+                else:
+                    # 3xx & 4xx errors
+                    response_data = {'success': False,
+                                     'code': response_code,
+                                     'result': str(response_data)}
+            elif response_type == 'text/html':
+                # used by media for preview
+                if response_code == requests.codes.ok:
+                    # 200 ok
+                    response_data = {'success': True,
+                                     'result': str(response_data)}
+                else:
+                    # 3xx & 4xx errors
+                    response_data = {'success': False,
+                                     'code': response_code,
+                                     'result': str(response_data)}
+
             else:
                 # Assuming nothing - but continuing anyway
                 # A single value is returned (vs an array or object)
                 if response_code == requests.codes.ok:
                     # 200 ok
-                    response_data = {'success': True, 'result': str(response_data)}
+                    response_data = {'success': True,
+                                     'result': str(response_data)}
                 else:
                     # 3xx & 4xx errors
-                    response_data = {'success': False, 'code': response_code, 'result': str(response_data)}
+                    response_data = {'success': False,
+                                     'code': response_code,
+                                     'result': str(response_data)}
 
             # it would be nice to return the error code and content type values; but not quite yet
             return response_data
@@ -303,22 +459,25 @@ class CloudFlare(object):
                     message = errors['error']
                 else:
                     message = ''
+                if 'messages' in response_data:
+                    errors['error_chain'] = response_data['messages']
                 if 'error_chain' in errors:
                     error_chain = errors['error_chain']
                     for error in error_chain:
                         if self.logger:
-                            self.logger.debug('Response: error %d %s - chain' %
-                                              (error['code'], error['message']))
+                            self.logger.debug('Response: error %d %s - chain',
+                                              error['code'],
+                                              error['message'])
                     if self.logger:
-                        self.logger.debug('Response: error %d %s' % (code, message))
+                        self.logger.debug('Response: error %d %s', code, message)
                     raise CloudFlareAPIError(code, message, error_chain)
                 else:
                     if self.logger:
-                        self.logger.debug('Response: error %d %s' % (code, message))
+                        self.logger.debug('Response: error %d %s', code, message)
                     raise CloudFlareAPIError(code, message)
 
             if self.logger:
-                self.logger.debug('Response: %s' % (response_data['result']))
+                self.logger.debug('Response: %s', response_data['result'])
             if self.raw:
                 result = {}
                 # theres always a result value
@@ -332,14 +491,27 @@ class CloudFlare(object):
                 result = response_data['result']
             return result
 
+        def _call_unwrapped(self, method, headers, parts,
+                            identifier1=None, identifier2=None, identifier3=None,
+                            params=None, data=None, files=None):
+            """ Cloudflare v4 API"""
+
+            response_data = self._raw(method, headers, parts,
+                                      identifier1, identifier2, identifier3,
+                                      params, data, files)
+            if self.logger:
+                self.logger.debug('Response: %s', response_data)
+            result = response_data
+            return result
+
     class _add_unused(object):
         """ Cloudflare v4 API"""
 
-        def __init__(self, base, api_call_part1, api_call_part2=None, api_call_part3=None):
+        def __init__(self, base, p1, p2=None, p3=None):
             """ Cloudflare v4 API"""
 
             self._base = base
-            self._parts = [api_call_part1, api_call_part2, api_call_part3]
+            self._parts_unused = [p1, p2, p3]
 
         def __call__(self, identifier1=None, identifier2=None, identifier3=None, params=None, data=None):
             """ Cloudflare v4 API"""
@@ -350,7 +522,7 @@ class CloudFlare(object):
         def __str__(self):
             """ Cloudflare v4 API"""
 
-            return '[%s]' % ('/' + '/:id/'.join(filter(None, self._parts)))
+            return '[%s]' % ('/' + '/:id/'.join(filter(None, self._parts_unused)))
 
         def get(self, identifier1=None, identifier2=None, identifier3=None, params=None, data=None):
             """ Cloudflare v4 API"""
@@ -362,7 +534,7 @@ class CloudFlare(object):
 
             raise CloudFlareAPIError(0, 'patch() call not available for this endpoint')
 
-        def post(self, identifier1=None, identifier2=None, identifier3=None, params=None, data=None):
+        def post(self, identifier1=None, identifier2=None, identifier3=None, params=None, data=None, files=None):
             """ Cloudflare v4 API"""
 
             raise CloudFlareAPIError(0, 'post() call not available for this endpoint')
@@ -380,11 +552,11 @@ class CloudFlare(object):
     class _add_noauth(object):
         """ Cloudflare v4 API"""
 
-        def __init__(self, base, api_call_part1, api_call_part2=None, api_call_part3=None):
+        def __init__(self, base, p1, p2=None, p3=None):
             """ Cloudflare v4 API"""
 
             self._base = base
-            self._parts = [api_call_part1, api_call_part2, api_call_part3]
+            self._parts = [p1, p2, p3]
 
         def __call__(self, identifier1=None, identifier2=None, identifier3=None, params=None, data=None):
             """ Cloudflare v4 API"""
@@ -409,7 +581,7 @@ class CloudFlare(object):
 
             raise CloudFlareAPIError(0, 'patch() call not available for this endpoint')
 
-        def post(self, identifier1=None, identifier2=None, identifier3=None, params=None, data=None):
+        def post(self, identifier1=None, identifier2=None, identifier3=None, params=None, data=None, files=None):
             """ Cloudflare v4 API"""
 
             raise CloudFlareAPIError(0, 'post() call not available for this endpoint')
@@ -427,11 +599,11 @@ class CloudFlare(object):
     class _add_with_auth(object):
         """ Cloudflare v4 API"""
 
-        def __init__(self, base, api_call_part1, api_call_part2=None, api_call_part3=None):
+        def __init__(self, base, p1, p2=None, p3=None):
             """ Cloudflare v4 API"""
 
             self._base = base
-            self._parts = [api_call_part1, api_call_part2, api_call_part3]
+            self._parts = [p1, p2, p3]
 
         def __call__(self, identifier1=None, identifier2=None, identifier3=None, params=None, data=None):
             """ Cloudflare v4 API"""
@@ -479,14 +651,69 @@ class CloudFlare(object):
                                              identifier1, identifier2, identifier3,
                                              params, data)
 
-    class _add_with_cert_auth(object):
+    class _add_with_auth_unwrapped(object):
         """ Cloudflare v4 API"""
 
-        def __init__(self, base, api_call_part1, api_call_part2=None, api_call_part3=None):
+        def __init__(self, base, p1, p2=None, p3=None):
             """ Cloudflare v4 API"""
 
             self._base = base
-            self._parts = [api_call_part1, api_call_part2, api_call_part3]
+            self._parts = [p1, p2, p3]
+
+        def __call__(self, identifier1=None, identifier2=None, identifier3=None, params=None, data=None):
+            """ Cloudflare v4 API"""
+
+            # This is the same as a get()
+            return self.get(identifier1, identifier2, identifier3, params, data)
+
+        def __str__(self):
+            """ Cloudflare v4 API"""
+
+            return '[%s]' % ('/' + '/:id/'.join(filter(None, self._parts)))
+
+        def get(self, identifier1=None, identifier2=None, identifier3=None, params=None, data=None):
+            """ Cloudflare v4 API"""
+
+            return self._base.call_with_auth_unwrapped('GET', self._parts,
+                                                       identifier1, identifier2, identifier3,
+                                                       params, data)
+
+        def patch(self, identifier1=None, identifier2=None, identifier3=None, params=None, data=None):
+            """ Cloudflare v4 API"""
+
+            return self._base.call_with_auth_unwrapped('PATCH', self._parts,
+                                                       identifier1, identifier2, identifier3,
+                                                       params, data)
+
+        def post(self, identifier1=None, identifier2=None, identifier3=None, params=None, data=None, files=None):
+            """ Cloudflare v4 API"""
+
+            return self._base.call_with_auth_unwrapped('POST', self._parts,
+                                                       identifier1, identifier2, identifier3,
+                                                       params, data, files)
+
+        def put(self, identifier1=None, identifier2=None, identifier3=None, params=None, data=None):
+            """ Cloudflare v4 API"""
+
+            return self._base.call_with_auth_unwrapped('PUT', self._parts,
+                                                       identifier1, identifier2, identifier3,
+                                                       params, data)
+
+        def delete(self, identifier1=None, identifier2=None, identifier3=None, params=None, data=None):
+            """ Cloudflare v4 API"""
+
+            return self._base.call_with_auth_unwrapped('DELETE', self._parts,
+                                                       identifier1, identifier2, identifier3,
+                                                       params, data)
+
+    class _add_with_cert_auth(object):
+        """ Cloudflare v4 API"""
+
+        def __init__(self, base, p1, p2=None, p3=None):
+            """ Cloudflare v4 API"""
+
+            self._base = base
+            self._parts = [p1, p2, p3]
 
         def __call__(self, identifier1=None, identifier2=None, identifier3=None, params=None, data=None):
             """ Cloudflare v4 API"""
@@ -534,6 +761,42 @@ class CloudFlare(object):
                                                  identifier1, identifier2, identifier3,
                                                  params, data)
 
+    def add(self, t, p1, p2=None, p3=None):
+        """add api call to class"""
+
+        a = []
+        if p1:
+            a += p1.split('/')
+        if p2:
+            a += p2.split('/')
+        if p3:
+            a += p3.split('/')
+
+        branch = self
+        for element in a[0:-1]:
+            try:
+                branch = getattr(branch, element)
+            except:
+                # should never happen
+                raise CloudFlareAPIError(0, 'api load name failed')
+        name = a[-1]
+
+        if t == 'VOID':
+            f = self._add_unused(self._base, p1, p2, p3)
+        elif t == 'OPEN':
+            f = self._add_noauth(self._base, p1, p2, p3)
+        elif t == 'AUTH':
+            f = self._add_with_auth(self._base, p1, p2, p3)
+        elif t == 'CERT':
+            f = self._add_with_cert_auth(self._base, p1, p2, p3)
+        elif t == 'AUTH_UNWRAPPED':
+            f = self._add_with_auth_unwrapped(self._base, p1, p2, p3)
+        else:
+            # should never happen
+            raise CloudFlareAPIError(0, 'api load type mismatch')
+
+        setattr(branch, name, f)
+
     def api_list(self, m=None, s=''):
         """recursive walk of the api tree returning a list of api calls"""
         if m is None:
@@ -557,7 +820,7 @@ class CloudFlare(object):
                 w = w + self.api_list(a, s + '/' + n)
         return w
 
-    def __init__(self, email=None, token=None, certtoken=None, debug=False, raw=False):
+    def __init__(self, email=None, token=None, certtoken=None, debug=False, raw=False, use_sessions=True):
         """ Cloudflare v4 API"""
 
         base_url = BASE_URL
@@ -572,7 +835,7 @@ class CloudFlare(object):
         if certtoken is None:
             certtoken = conf_certtoken
 
-        self._base = self._v4base(email, token, certtoken, base_url, debug, raw)
+        self._base = self._v4base(email, token, certtoken, base_url, debug, raw, use_sessions)
 
         # add the API calls
         api_v4(self)
